@@ -1,4 +1,4 @@
-/*  $Id: xml_format.c,v 1.20 2004/02/12 23:28:26 mgrouch Exp $  */
+/*  $Id: xml_format.c,v 1.25 2005/01/07 02:33:40 mgrouch Exp $  */
 
 /*
 
@@ -43,7 +43,9 @@ THE SOFTWARE.
 #include <libxml/parserInternals.h>
 #include <libxml/uri.h>
 
+#ifndef HAVE_STRDUP
 #include "strdup.h"
+#endif
 
 /*
  *  TODO:  1. Attribute formatting options (as every attribute on a new line)
@@ -66,7 +68,7 @@ typedef struct _foOptions {
 typedef foOptions *foOptionsPtr;
 
 /*
- * usage string chunk : 509 char min on ISO C90
+ * usage string chunk : 509 char max on ISO C90
  */
 static const char format_usage_str_1[] =
 "XMLStarlet Toolkit: Format XML document\n"
@@ -98,9 +100,9 @@ foUsage(int argc, char **argv)
 {
     extern const char more_info[];
     FILE* o = stderr;
-    fprintf(o, format_usage_str_1);
-    fprintf(o, format_usage_str_2);
-    fprintf(o, more_info);
+    fprintf(o, "%s", format_usage_str_1);
+    fprintf(o, "%s", format_usage_str_2);
+    fprintf(o, "%s", more_info);
     exit(1);
 }
 
@@ -152,7 +154,11 @@ foInitLibXml(foOptionsPtr ops)
     if (ops->indent)
     {
         xmlIndentTreeOutput = 1;
-        if (ops->indent_spaces > 0)
+        if (ops->indent_tab) 
+        {
+            xmlTreeIndentString = strdup("\t");
+        }
+        else if (ops->indent_spaces > 0)
         {
             int i;
             char *p;
@@ -161,7 +167,6 @@ foInitLibXml(foOptionsPtr ops)
             for (i=0; i<ops->indent_spaces; i++) p[i] = ' ';
             p[ops->indent_spaces] = 0;
         }
-        if (ops->indent_tab) xmlTreeIndentString = strdup("\t");
     }
     else
         xmlIndentTreeOutput = 0;
@@ -348,6 +353,8 @@ foProcess(foOptionsPtr ops, int start, int argc, char **argv)
         }
         ret = xmlOutputBufferClose(buf);
     }
+    
+    xmlFreeDoc(doc);
     return ret;
 }
 
@@ -357,8 +364,10 @@ foProcess(foOptionsPtr ops, int start, int argc, char **argv)
 void
 foCleanup()
 {
-    xmlCleanupParser();
     if (xmlTreeIndentString) free((char *)xmlTreeIndentString);
+    xmlTreeIndentString = NULL;
+    
+    xmlCleanupParser();
 #if 0
     xmlMemoryDump();
 #endif
