@@ -1,4 +1,4 @@
-/* $Id: xml_pyx.c,v 1.6 2004/11/11 03:39:34 mgrouch Exp $ */
+/* $Id: xml_pyx.c,v 1.9 2005/03/12 03:24:23 mgrouch Exp $ */
 
 /**
  *  Based on xmln from pyxie project
@@ -86,19 +86,19 @@ pyxStartElement(void *userData, const xmlChar *name, const xmlChar **atts)
         for (p = atts; *p != NULL; p++);
 
         AttributeCount = (p - atts) >> 1; /* (name,value) pairs 
-					     so divide by two */
+                                           so divide by two */
         if (AttributeCount > 1)
             /* Sort the pairs based on the name part of the pair */
             qsort ((void *)atts,
-		    AttributeCount,
-		    sizeof(char *)*2,
-		    CompareAttributes);
+                    AttributeCount,
+                    sizeof(char *)*2,
+                    CompareAttributes);
 
         while (*atts) {
             /* Attribute Name */
             fprintf (stdout,"A%s ",*atts);
             atts++; /* Now pointing at value - can contain literal "\n" 
-		       so escape */
+                       so escape */
             SanitizeData((const char *) *atts, xmlStrlen(*atts));
             atts++;
             putchar('\n');
@@ -109,21 +109,21 @@ pyxStartElement(void *userData, const xmlChar *name, const xmlChar **atts)
 void
 pyxEndElement(void *userData, const xmlChar *name)
 {
-    fprintf (stdout,")%s\n",name);
+    fprintf(stdout,")%s\n",name);
 }
 
 void
 pyxCharacterData(void *userData, const xmlChar *s, int len)
 {
-    fprintf (stdout, "-");
-    SanitizeData ((const char *) s,len);
-    putchar ('\n');
+    fprintf(stdout, "-");
+    SanitizeData((const char *) s,len);
+    putchar('\n');
 }
 
 void
 pyxProcessingInstruction(void *userData, 
                          const xmlChar *target, 
-			 const xmlChar *data)
+                         const xmlChar *data)
 {
     fprintf(stdout,"?%s ",target);
     SanitizeData((const char *) data, xmlStrlen(data));
@@ -151,7 +151,7 @@ pyxNotationDeclHandler(void *userData,
 {
     fprintf(stdout, "N%s %s%s%s\n", (char*) notationName, (char*) systemId,
            (publicId == NULL? "": " "), 
-	   (publicId == NULL? "": (const char*) publicId));
+           (publicId == NULL? "": (const char*) publicId));
 }
 
 void
@@ -167,6 +167,37 @@ pyxExternalEntityReferenceHandler(void* userData,
     }
 }
 
+static void
+pyxExternalSubsetHandler(void *ctx ATTRIBUTE_UNUSED, const xmlChar *name,
+                         const xmlChar *ExternalID, const xmlChar *SystemID)
+{
+    fprintf(stdout, "D %s PUBLIC", name); /* TODO: re-check */
+    if (ExternalID == NULL)
+        fprintf(stdout, " ");
+    else
+        fprintf(stdout, " \"%s\"", ExternalID);
+    if (SystemID == NULL)
+        fprintf(stdout, "\n");
+    else
+        fprintf(stdout, " \"%s\"\n", SystemID);
+}
+
+static void
+pyxCommentHandler(void *ctx ATTRIBUTE_UNUSED, const xmlChar *value)
+{
+    fprintf(stdout,"C");
+    SanitizeData((const char *) value, xmlStrlen(value));
+    fprintf(stdout,"\n");
+}
+
+static void
+pyxCdataBlockHandler(void *ctx ATTRIBUTE_UNUSED, const xmlChar *value, int len)
+{
+    fprintf(stdout,"[");
+    SanitizeData((const char *) value, len);
+    fprintf(stdout,"\n");
+}
+
 static void
 pyxUsage()
 {
@@ -198,6 +229,9 @@ pyx_process_file(const char *filename)
     xmlSAX_handler.notationDecl = pyxNotationDeclHandler;
     xmlSAX_handler.reference = pyxExternalEntityReferenceHandler;
     xmlSAX_handler.unparsedEntityDecl = pyxUnparsedEntityDeclHandler;
+    xmlSAX_handler.externalSubset = pyxExternalSubsetHandler;
+    xmlSAX_handler.comment = pyxCommentHandler;
+    xmlSAX_handler.cdataBlock = pyxCdataBlockHandler;
 
     ret = xmlSAXUserParseFile(&xmlSAX_handler, NULL, filename);
     xmlCleanupParser();
