@@ -1,10 +1,10 @@
-/*  $Id: xml_select.c,v 1.61 2004/03/30 04:14:40 mgrouch Exp $  */
+/*  $Id: xml_select.c,v 1.67 2005/01/07 02:02:13 mgrouch Exp $  */
 
 /*
 
 XMLStarlet: Command Line Toolkit to query/edit/check/transform XML documents
 
-Copyright (c) 2002 Mikhail Grushinskiy.  All Rights Reserved.
+Copyright (c) 2002-2004 Mikhail Grushinskiy.  All Rights Reserved.
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
@@ -55,12 +55,13 @@ typedef struct _selOptions {
     int noblanks;             /* Remove insignificant spaces from XML tree */
     int no_omit_decl;         /* Print XML declaration line <?xml version="1.0"?> */
     int nonet;                /* refuse to fetch DTDs or entities over network */
+    const char *encoding;     /* the "encoding" attribute on the stylesheet's <xsl:output/> */
 } selOptions;
 
 typedef selOptions *selOptionsPtr;
 
 /*
- * usage string chunk : 509 char min on ISO C90
+ * usage string chunk : 509 char max on ISO C90
  */
 static const char select_usage_str_1[] =
 "XMLStarlet Toolkit: Select from XML document(s)\n"
@@ -72,36 +73,37 @@ static const char select_usage_str_1[] =
 
 static const char select_usage_str_2[] =
 "<global-options> are:\n"
-"  -C or --comp       - display generated XSLT\n"
-"  -R or --root       - print root element <xsl-select>\n"
-"  -T or --text       - output is text (default is XML)\n"
-"  -I or --indent     - indent output\n"
-"  -D or --xml-decl   - do not omit xml declaration line\n"
-"  -B or --noblanks   - remove insignificant spaces from XML tree\n"
-"  -N <name>=<value>  - predefine namespaces (name without \'xmlns:\')\n";
+"  -C or --comp              - display generated XSLT\n"
+"  -R or --root              - print root element <xsl-select>\n"
+"  -T or --text              - output is text (default is XML)\n"
+"  -I or --indent            - indent output\n"
+"  -D or --xml-decl          - do not omit xml declaration line\n"
+"  -B or --noblanks          - remove insignificant spaces from XML tree\n"
+"  -E or --encode <encoding> - output in the given encoding (utf-8, unicode...)\n";
 
 static const char select_usage_str_3[] =
-"                       ex: xsql=urn:oracle-xsql\n"
-"                       Multiple -N options are allowed.\n"
-"  --net              - allow fetch DTDs or entities over network\n"
-"  --help             - display help\n\n";
+"  -N <name>=<value>         - predefine namespaces (name without \'xmlns:\')\n"
+"                              ex: xsql=urn:oracle-xsql\n"
+"                              Multiple -N options are allowed.\n"
+"  --net                     - allow fetch DTDs or entities over network\n"
+"  --help                    - display help\n\n";
 
 static const char select_usage_str_4[] =
 "Syntax for templates: -t|--template <options>\n"
 "where <options>\n"
-"  -c or --copy-of <xpath>  - print copy of XPATH expression\n"
-"  -v or --value-of <xpath> - print value of XPATH expression\n"
-"  -o or --output <string>  - output string literal\n"
-"  -n or --nl               - print new line\n"
-"  -f or --inp-name         - print input file name (or URL)\n";
+"  -c or --copy-of <xpath>   - print copy of XPATH expression\n"
+"  -v or --value-of <xpath>  - print value of XPATH expression\n"
+"  -o or --output <string>   - output string literal\n"
+"  -n or --nl                - print new line\n"
+"  -f or --inp-name          - print input file name (or URL)\n";
 
 static const char select_usage_str_5[] =
-"  -m or --match <xpath>    - match XPATH expression\n"
-"  -i or --if <test-xpath>  - check condition <xsl:if test=\"test-xpath\">\n"
-"  -e or --elem <name>      - print out element <xsl:element name=\"name\">\n"
-"  -a or --attr <name>      - add attribute <xsl:attribute name=\"name\">\n"
-"  -b or --break            - break nesting\n"
-"  -s or --sort op xpath    - sort in order (used after -m) where\n";
+"  -m or --match <xpath>     - match XPATH expression\n"
+"  -i or --if <test-xpath>   - check condition <xsl:if test=\"test-xpath\">\n"
+"  -e or --elem <name>       - print out element <xsl:element name=\"name\">\n"
+"  -a or --attr <name>       - add attribute <xsl:attribute name=\"name\">\n"
+"  -b or --break             - break nesting\n"
+"  -s or --sort op xpath     - sort in order (used after -m) where\n";
 
 static const char select_usage_str_6[] =
 "  op is X:Y:Z, \n"
@@ -155,17 +157,17 @@ selUsage(int argc, char **argv)
     extern const char more_info[];
     extern const char libxslt_more_info[];
     FILE* o = stderr;
-    fprintf(o, select_usage_str_1);
-    fprintf(o, select_usage_str_2);
-    fprintf(o, select_usage_str_3);
-    fprintf(o, select_usage_str_4);
-    fprintf(o, select_usage_str_5);
-    fprintf(o, select_usage_str_6);
-    fprintf(o, select_usage_str_7);
-    fprintf(o, select_usage_str_8);
-    fprintf(o, select_usage_str_9);
-    fprintf(o, more_info);
-    fprintf(o, libxslt_more_info);
+    fprintf(o, "%s", select_usage_str_1);
+    fprintf(o, "%s", select_usage_str_2);
+    fprintf(o, "%s", select_usage_str_3);
+    fprintf(o, "%s", select_usage_str_4);
+    fprintf(o, "%s", select_usage_str_5);
+    fprintf(o, "%s", select_usage_str_6);
+    fprintf(o, "%s", select_usage_str_7);
+    fprintf(o, "%s", select_usage_str_8);
+    fprintf(o, "%s", select_usage_str_9);
+    fprintf(o, "%s", more_info);
+    fprintf(o, "%s", libxslt_more_info);
     exit(1);
 }
 
@@ -182,6 +184,7 @@ selInitOptions(selOptionsPtr ops)
     ops->noblanks = 0;
     ops->no_omit_decl = 0;
     ops->nonet = 1;
+    ops->encoding = NULL;
 }
 
 /**
@@ -286,6 +289,27 @@ selParseOptions(selOptionsPtr ops, int argc, char **argv)
         {
             ops->no_omit_decl = 1;
         }
+        else if (!strcmp(argv[i], "-E") || !strcmp(argv[i], "--encode"))
+        {
+            if ((i+1) < argc)
+            {
+                if (argv[i + 1][0] == '-')
+                {
+                    fprintf(stderr, "-E option requires argument <encoding> ex: (utf-8, unicode...)\n");
+                    exit(2);
+                }
+                else
+                {
+                    ops->encoding = argv[i + 1];
+                }
+            }
+            else
+            {
+                fprintf(stderr, "-E option requires argument <encoding> ex: (utf-8, unicode...)\n");
+                exit(2);
+            }
+
+        }
         else if (!strcmp(argv[i], "--net"))
         {
             ops->nonet = 0;
@@ -334,10 +358,10 @@ selGenTemplate(char* xsl_buf, int *len, selOptionsPtr ops, int num,
     }
 
     templateEmpty = 1;
-    c += sprintf(xsl_buf + c, "<xsl:template name=\"t%d\">\n", t);
+    c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "<xsl:template name=\"t%d\">\n", t);
 
     stack = stack_create(max_depth);
-    
+
     i++;
     m = 0;
     while(i < argc)
@@ -345,8 +369,8 @@ selGenTemplate(char* xsl_buf, int *len, selOptionsPtr ops, int num,
         if(!strcmp(argv[i], "-c") || !strcmp(argv[i], "--copy-of"))
         {
             templateEmpty = 0;
-            for (j=0; j <= m; j++) c += sprintf(xsl_buf + c, "  ");
-            c += sprintf(xsl_buf + c, "<xsl:copy-of select=\"%s\"/>\n", argv[i+1]);
+            for (j=0; j <= m; j++) c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "  ");
+            c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "<xsl:copy-of select=\"%s\"/>\n", argv[i+1]);
             i++;
         }
         else if(!strcmp(argv[i], "-v") || !strcmp(argv[i], "--value-of"))
@@ -358,8 +382,8 @@ selGenTemplate(char* xsl_buf, int *len, selOptionsPtr ops, int num,
                 stack_free(stack);
                 exit (1);
             }
-            for (j=0; j <= m; j++) c += sprintf(xsl_buf + c, "  ");
-            c += sprintf(xsl_buf + c, "<xsl:value-of select=\"%s\"/>\n", argv[i+1]);
+            for (j=0; j <= m; j++) c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "  ");
+            c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "<xsl:value-of select=\"%s\"/>\n", argv[i+1]);
             i++;
         }
         else if(!strcmp(argv[i], "-o") || !strcmp(argv[i], "--output"))
@@ -371,21 +395,21 @@ selGenTemplate(char* xsl_buf, int *len, selOptionsPtr ops, int num,
                 stack_free(stack);
                 exit (1);
             }
-            for (j=0; j <= m; j++) c += sprintf(xsl_buf + c, "  ");
-            c += sprintf(xsl_buf + c, "<xsl:value-of select=\"'%s'\"/>\n", argv[i+1]);
+            for (j=0; j <= m; j++) c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "  ");
+            c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "<xsl:value-of select=\"'%s'\"/>\n", argv[i+1]);
             i++;
         }
         else if(!strcmp(argv[i], "-f") || !strcmp(argv[i], "--inp-name"))
         {
             templateEmpty = 0;
-            for (j=0; j <= m; j++) c += sprintf(xsl_buf + c, "  ");
-            c += sprintf(xsl_buf + c, "<xsl:copy-of select=\"$inputFile\"/>\n");
+            for (j=0; j <= m; j++) c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "  ");
+            c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "<xsl:copy-of select=\"$inputFile\"/>\n");
         }
         else if(!strcmp(argv[i], "-n") || !strcmp(argv[i], "--nl"))
         {
             templateEmpty = 0;
-            for (j=0; j <= m; j++) c += sprintf(xsl_buf + c, "  ");
-            c += sprintf(xsl_buf + c, "<xsl:value-of select=\"'&#10;'\"/>\n");
+            for (j=0; j <= m; j++) c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "  ");
+            c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "<xsl:value-of select=\"'&#10;'\"/>\n");
         }
         else if(!strcmp(argv[i], "-i") || !strcmp(argv[i], "--if"))
         {
@@ -396,8 +420,8 @@ selGenTemplate(char* xsl_buf, int *len, selOptionsPtr ops, int num,
                 stack_free(stack);
                 exit (1);
             }
-            for (j=0; j <= m; j++) c += sprintf(xsl_buf + c, "  ");
-            c += sprintf(xsl_buf + c, "<xsl:if test=\"%s\">\n", argv[i+1]);
+            for (j=0; j <= m; j++) c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "  ");
+            c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "<xsl:if test=\"%s\">\n", argv[i+1]);
             stack_push(stack, STK_IF);
             m++;
             i++;
@@ -411,8 +435,8 @@ selGenTemplate(char* xsl_buf, int *len, selOptionsPtr ops, int num,
                 stack_free(stack);
                 exit (1);
             }
-            for (j=0; j <= m; j++) c += sprintf(xsl_buf + c, "  ");
-            c += sprintf(xsl_buf + c, "<xsl:element name=\"%s\">\n", argv[i+1]);
+            for (j=0; j <= m; j++) c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "  ");
+            c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "<xsl:element name=\"%s\">\n", argv[i+1]);
             stack_push(stack, STK_ELEM);
             m++;
             i++;
@@ -426,8 +450,8 @@ selGenTemplate(char* xsl_buf, int *len, selOptionsPtr ops, int num,
                 stack_free(stack);
                 exit (1);
             }
-            for (j=0; j <= m; j++) c += sprintf(xsl_buf + c, "  ");
-            c += sprintf(xsl_buf + c, "<xsl:attribute name=\"%s\">\n", argv[i+1]);
+            for (j=0; j <= m; j++) c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "  ");
+            c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "<xsl:attribute name=\"%s\">\n", argv[i+1]);
             stack_push(stack, STK_ATTR);
             m++;
             i++;
@@ -441,8 +465,8 @@ selGenTemplate(char* xsl_buf, int *len, selOptionsPtr ops, int num,
                 stack_free(stack);
                 exit (1);
             }
-            for (j=0; j <= m; j++) c += sprintf(xsl_buf + c, "  ");
-            c += sprintf(xsl_buf + c, "<xsl:for-each select=\"%s\">\n", argv[i+1]);
+            for (j=0; j <= m; j++) c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "  ");
+            c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "<xsl:for-each select=\"%s\">\n", argv[i+1]);
             stack_push(stack, STK_MATCH);
             m++;
             i++;
@@ -471,21 +495,21 @@ selGenTemplate(char* xsl_buf, int *len, selOptionsPtr ops, int num,
                 }
                 i++;
                 Select=argv[i];
-                
-                for (j=0; j <= m; j++) c += sprintf(xsl_buf + c, "  ");
 
-                if (Order == 'A') c += sprintf(xsl_buf + c, "<xsl:sort order=\"ascending\"");
-                else c += sprintf(xsl_buf + c, "<xsl:sort order=\"descending\"");
+                for (j=0; j <= m; j++) c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "  ");
 
-                if (Type == 'N') c += sprintf(xsl_buf + c, " data-type=\"number\"");
-                else c += sprintf(xsl_buf + c, " data-type=\"text\"");
+                if (Order == 'A') c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "<xsl:sort order=\"ascending\"");
+                else c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "<xsl:sort order=\"descending\"");
 
-                if (Case == 'L') c += sprintf(xsl_buf + c, " case-order=\"lower-first\"");
-                else c += sprintf(xsl_buf + c, " case-order=\"upper-first\"");
+                if (Type == 'N') c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, " data-type=\"number\"");
+                else c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, " data-type=\"text\"");
 
-                c += sprintf(xsl_buf + c, " select=\"%s\"", Select);
-                
-                c += sprintf(xsl_buf + c, "/>\n");
+                if (Case == 'L') c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, " case-order=\"lower-first\"");
+                else c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, " case-order=\"upper-first\"");
+
+                c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, " select=\"%s\"", Select);
+
+                c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "/>\n");
             }
         }
         else if(!strcmp(argv[i], "-t") || !strcmp(argv[i], "--template"))
@@ -500,26 +524,26 @@ selGenTemplate(char* xsl_buf, int *len, selOptionsPtr ops, int num,
             {
                 StackItem itm;
 
-                for (j=0; j<stack_depth(stack); j++) c += sprintf(xsl_buf + c, "  ");
+                for (j=0; j<stack_depth(stack); j++) c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "  ");
                 itm = stack_pop(stack);
                 if (itm == STK_MATCH)
                 {
-                    c += sprintf(xsl_buf + c, "</xsl:for-each>\n");
+                    c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "</xsl:for-each>\n");
                     m--;
                 }
                 else if (itm == STK_IF)
                 {
-                    c += sprintf(xsl_buf + c, "</xsl:if>\n");
+                    c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "</xsl:if>\n");
                     m--;
                 }
                 else if (itm == STK_ELEM)
                 {
-                    c += sprintf(xsl_buf + c, "</xsl:element>\n");
+                    c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "</xsl:element>\n");
                     m--;
                 }
                 else if (itm == STK_ATTR)
                 {
-                    c += sprintf(xsl_buf + c, "</xsl:attribute>\n");
+                    c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "</xsl:attribute>\n");
                     m--;
                 }
                 /* printf("%c\n", itm); */
@@ -548,8 +572,8 @@ selGenTemplate(char* xsl_buf, int *len, selOptionsPtr ops, int num,
 /*
     if((i < argc) && (!strcmp(argv[i], "-n") || !strcmp(argv[i], "--nl")))
     {
-        for (j=0; j <= m; j++) c += sprintf(xsl_buf + c, "  ");
-        c += sprintf(xsl_buf + c, "<xsl:value-of select=\"'&#10;'\"/>\n");
+        for (j=0; j <= m; j++) c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "  ");
+        c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "<xsl:value-of select=\"'&#10;'\"/>\n");
     }
 */
 
@@ -557,19 +581,19 @@ selGenTemplate(char* xsl_buf, int *len, selOptionsPtr ops, int num,
     {
         StackItem itm;
 
-        for (j=0; j<stack_depth(stack); j++) c += sprintf(xsl_buf + c, "  ");
+        for (j=0; j<stack_depth(stack); j++) c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "  ");
         itm = stack_pop(stack);
-        if (itm == STK_MATCH) c += sprintf(xsl_buf + c, "</xsl:for-each>\n");
-        else if (itm == STK_IF) c += sprintf(xsl_buf + c, "</xsl:if>\n");
-        else if (itm == STK_ELEM) c += sprintf(xsl_buf + c, "</xsl:element>\n");
-        else if (itm == STK_ATTR) c += sprintf(xsl_buf + c, "</xsl:attribute>\n");
+        if (itm == STK_MATCH) c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "</xsl:for-each>\n");
+        else if (itm == STK_IF) c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "</xsl:if>\n");
+        else if (itm == STK_ELEM) c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "</xsl:element>\n");
+        else if (itm == STK_ATTR) c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "</xsl:attribute>\n");
 /*        printf("%c\n", itm);  */
     }
 /*
     for (k=0; k<m; k++)
     {
-        for (j=k; j<m; j++) c += sprintf(xsl_buf + c, "  ");
-        c += sprintf(xsl_buf + c, "</xsl:for-each>\n");
+        for (j=k; j<m; j++) c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "  ");
+        c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "</xsl:for-each>\n");
     }
 */
     if (templateEmpty)
@@ -581,7 +605,7 @@ selGenTemplate(char* xsl_buf, int *len, selOptionsPtr ops, int num,
         exit(3);
     }
 
-    c += sprintf(xsl_buf + c, "</xsl:template>\n");
+    c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "</xsl:template>\n");
 
     *len = c;
 
@@ -615,59 +639,61 @@ selGenTemplate(char* xsl_buf, int *len, selOptionsPtr ops, int num,
  *  Prepare XSLT stylesheet based on command line options
  */
 int
-selPrepareXslt(char* xsl_buf, int *len, selOptionsPtr ops, const char *ns_arr[], 
+selPrepareXslt(char* xsl_buf, int *len, selOptionsPtr ops, const char *ns_arr[],
                int start, int argc, char **argv)
 {
     int c, i, t, ns;
-   
+
     xsl_buf[0] = 0;
     *len = 0;
     c = 0;
 
     c += sprintf(xsl_buf, "<?xml version=\"1.0\"?>\n");
-    c += sprintf(xsl_buf + c,
+    c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1,
       "<xsl:stylesheet version=\"1.0\" xmlns:xsl=\"http://www.w3.org/1999/XSL/Transform\"");
-    c += sprintf(xsl_buf + c, "\n xmlns:exslt=\"http://exslt.org/common\"");
-    c += sprintf(xsl_buf + c, "\n xmlns:math=\"http://exslt.org/math\"");
-    c += sprintf(xsl_buf + c, "\n xmlns:date=\"http://exslt.org/dates-and-times\"");
-    c += sprintf(xsl_buf + c, "\n xmlns:func=\"http://exslt.org/functions\"");
-    c += sprintf(xsl_buf + c, "\n xmlns:set=\"http://exslt.org/sets\"");
-    c += sprintf(xsl_buf + c, "\n xmlns:str=\"http://exslt.org/strings\"");
-    c += sprintf(xsl_buf + c, "\n xmlns:dyn=\"http://exslt.org/dynamic\"");
-    c += sprintf(xsl_buf + c, "\n xmlns:saxon=\"http://icl.com/saxon\"");
-    c += sprintf(xsl_buf + c, "\n xmlns:xalanredirect=\"org.apache.xalan.xslt.extensions.Redirect\"");
-    c += sprintf(xsl_buf + c, "\n xmlns:xt=\"http://www.jclark.com/xt\"");
-    c += sprintf(xsl_buf + c, "\n xmlns:libxslt=\"http://xmlsoft.org/XSLT/namespace\"");
-    c += sprintf(xsl_buf + c, "\n xmlns:test=\"http://xmlsoft.org/XSLT/\"");
+    c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "\n xmlns:exslt=\"http://exslt.org/common\"");
+    c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "\n xmlns:math=\"http://exslt.org/math\"");
+    c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "\n xmlns:date=\"http://exslt.org/dates-and-times\"");
+    c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "\n xmlns:func=\"http://exslt.org/functions\"");
+    c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "\n xmlns:set=\"http://exslt.org/sets\"");
+    c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "\n xmlns:str=\"http://exslt.org/strings\"");
+    c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "\n xmlns:dyn=\"http://exslt.org/dynamic\"");
+    c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "\n xmlns:saxon=\"http://icl.com/saxon\"");
+    c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "\n xmlns:xalanredirect=\"org.apache.xalan.xslt.extensions.Redirect\"");
+    c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "\n xmlns:xt=\"http://www.jclark.com/xt\"");
+    c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "\n xmlns:libxslt=\"http://xmlsoft.org/XSLT/namespace\"");
+    c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "\n xmlns:test=\"http://xmlsoft.org/XSLT/\"");
 
     ns = 0;
     while(ns_arr[ns])
     {
         if (strlen(ns_arr[ns]))
-           c += sprintf(xsl_buf + c, "\n xmlns:%s=\"%s\"", ns_arr[ns], ns_arr[ns+1]);
+           c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "\n xmlns:%s=\"%s\"", ns_arr[ns], ns_arr[ns+1]);
         else
-           c += sprintf(xsl_buf + c, "\n xmlns=\"%s\"", ns_arr[ns+1]);
+           c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "\n xmlns=\"%s\"", ns_arr[ns+1]);
         ns += 2;
     }
     selCleanupNSArr(ns_arr);
-    
-    c += sprintf(xsl_buf + c, "\n extension-element-prefixes=\"exslt math date func set str dyn saxon xalanredirect xt libxslt test\"");
-    c += sprintf(xsl_buf + c, "\n exclude-result-prefixes=\"math str\"");
 
-    
-    c += sprintf(xsl_buf + c, ">\n");
+    c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1,
+                  "\n extension-element-prefixes=\"exslt math date func set str dyn saxon xalanredirect xt libxslt test\"");
+    c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "\n exclude-result-prefixes=\"math str\"");
 
-    if (ops->no_omit_decl) c += sprintf(xsl_buf + c, "<xsl:output omit-xml-declaration=\"no\"");
-    else c += sprintf(xsl_buf + c, "<xsl:output omit-xml-declaration=\"yes\"");
-    if (ops->indent) c += sprintf(xsl_buf + c, " indent=\"yes\"");
-    else c += sprintf(xsl_buf + c, " indent=\"no\"");
-    if (ops->outText) c += sprintf(xsl_buf + c, " method=\"text\"");
-    c += sprintf(xsl_buf + c, "/>\n");
 
-    c += sprintf(xsl_buf + c, "<xsl:param name=\"inputFile\">-</xsl:param>\n");
+    c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, ">\n");
 
-    c += sprintf(xsl_buf + c, "<xsl:template match=\"/\">\n");
-    if (!ops->outText && ops->printRoot) c += sprintf(xsl_buf + c, "<xml-select>\n");
+    if (ops->no_omit_decl) c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "<xsl:output omit-xml-declaration=\"no\"");
+    else c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "<xsl:output omit-xml-declaration=\"yes\"");
+    if (ops->indent) c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, " indent=\"yes\"");
+    else c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, " indent=\"no\"");
+    if (ops->encoding != NULL) c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, " encoding=\"%s\"", ops->encoding);
+    if (ops->outText) c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, " method=\"text\"");
+    c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "/>\n");
+
+    c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "<xsl:param name=\"inputFile\">-</xsl:param>\n");
+
+    c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "<xsl:template match=\"/\">\n");
+    if (!ops->outText && ops->printRoot) c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "<xml-select>\n");
 
     t = 0;
     i = start;
@@ -676,12 +702,12 @@ selPrepareXslt(char* xsl_buf, int *len, selOptionsPtr ops, const char *ns_arr[],
         if(!strcmp(argv[i], "-t") || !strcmp(argv[i], "--template"))
         {
             t++;
-            c += sprintf(xsl_buf + c, "  <xsl:call-template name=\"t%d\"/>\n", t);
+            c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "  <xsl:call-template name=\"t%d\"/>\n", t);
         }
         i++;
     }
-    if (!ops->outText && ops->printRoot) c += sprintf(xsl_buf + c, "</xml-select>\n");
-    c += sprintf(xsl_buf + c, "</xsl:template>\n");
+    if (!ops->outText && ops->printRoot) c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "</xml-select>\n");
+    c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "</xsl:template>\n");
 
     /*
      *  At least one -t option must be found
@@ -692,7 +718,7 @@ selPrepareXslt(char* xsl_buf, int *len, selOptionsPtr ops, const char *ns_arr[],
         fprintf(stderr, " no -t or --template options found\n");
         exit(2);
     }
-        
+
     t = 0;
     i = start;
     while(i < argc)
@@ -706,7 +732,7 @@ selPrepareXslt(char* xsl_buf, int *len, selOptionsPtr ops, const char *ns_arr[],
         }
     }
 
-    c += sprintf(xsl_buf + c, "</xsl:stylesheet>\n");
+    c += snprintf(xsl_buf + c, MAX_XSL_BUF - c - 1, "</xsl:stylesheet>\n");
     *len = c;
 
     return i;
@@ -725,19 +751,19 @@ selMain(int argc, char **argv)
     int start, c, i, n;
     int nCount = 0;
     int nbparams;
-  
+
     if (argc <= 2) selUsage(argc, argv);
 
     selInitOptions(&ops);
     xsltInitOptions(&xsltOps);
     start = selParseOptions(&ops, argc, argv);
     xsltOps.nonet = ops.nonet;
-    xsltOps.noblanks = ops.noblanks;    
+    xsltOps.noblanks = ops.noblanks;
     xsltInitLibXml(&xsltOps);
 
     /* set parameters */
     selParseNSArr(ns_arr, &nCount, argc-start, argv+2);
-    
+
     c = sizeof(xsl_buf);
     i = selPrepareXslt(xsl_buf, &c, &ops, ns_arr, start, argc, argv);
 
@@ -746,7 +772,7 @@ selMain(int argc, char **argv)
         fprintf(stdout, "%s", xsl_buf);
         exit(0);
     }
-    
+
     for (n=i; n<argc; n++)
     {
         xmlChar *value;
@@ -770,12 +796,16 @@ selMain(int argc, char **argv)
             xmlDocPtr doc = NULL;
             if (!cur) exit(2);
             doc = xmlParseFile(argv[n]);
-            xsltProcess(&xsltOps, doc, params, cur, argv[n]);
+            if (doc != NULL) {
+                xsltProcess(&xsltOps, doc, params, cur, argv[n]);
+            }
             xsltFreeStylesheet(cur);
         }
+        
+        xmlFree(value);
     }
 
-    if (i == argc)    
+    if (i == argc)
     {
         /*
          *  Parse XSLT stylesheet
@@ -785,9 +815,18 @@ selMain(int argc, char **argv)
         xmlDocPtr doc = NULL;
         if (!cur) exit(2);
         doc = xmlParseFile("-");
-        xsltProcess(&xsltOps, doc, params, cur, "-");
-        /* xsltFreeStylesheet(cur); */
+        if (doc != NULL) {
+            xsltProcess(&xsltOps, doc, params, cur, "-");
+        }
+        xsltFreeStylesheet(cur);
     }
-        
+
+    /* 
+     * Shutdown libxml
+     */
+    xsltCleanupGlobals();
+    xmlCleanupParser();
+    
     return 0;
-}  
+}
+
